"""
Filesystem utilities for managing map folding computation results.

This module provides functions for standardized handling of files related to the mapFolding
package, with a focus on saving, retrieving, and naming computation results. It implements
consistent naming conventions and path resolution strategies to ensure that:

1. Computation results are stored in a predictable location
2. Filenames follow a consistent pattern based on map dimensions
3. Results can be reliably retrieved for future reference
4. The system handles file operations safely with appropriate error handling

The module serves as the interface between the computational components of the package
and the filesystem, abstracting away the details of file operations and path management.
"""
from mapFolding.theSSOT import The
from os import PathLike
from pathlib import Path, PurePath
from sys import modules as sysModules
from typing import Any
import os

def getFilenameFoldsTotal(mapShape: tuple[int, ...]) -> str:
	"""
	Create a standardized filename for a computed `foldsTotal` value.

	This function generates a consistent, filesystem-safe filename based on map dimensions.
	Standardizing filenames ensures that results can be reliably stored and retrieved,
	avoiding potential filesystem incompatibilities or Python naming restrictions.

	Parameters:
		mapShape: A sequence of integers representing the dimensions of the map.

	Returns:
		filenameFoldsTotal: A filename string in format 'pMxN.foldsTotal' where M,N are sorted dimensions.

	Notes:
		The filename format ensures:
		- No spaces in the filename
		- Safe filesystem characters
		- Unique extension (.foldsTotal)
		- Python-safe strings (no starting with numbers, no reserved words)
		- The 'p' prefix preserves compatibility with Lunnan's original code.
	"""
	return 'p' + 'x'.join(str(dimension) for dimension in sorted(mapShape)) + '.foldsTotal'

def getPathFilenameFoldsTotal(mapShape: tuple[int, ...], pathLikeWriteFoldsTotal: PathLike[str] | PurePath | None = None) -> Path:
	"""
	Get a standardized path and filename for the computed foldsTotal value.

	This function resolves paths for storing computation results, handling different
	input types including directories, absolute paths, or relative paths. It ensures
	that all parent directories exist in the resulting path.

	Parameters:
		mapShape: List of dimensions for the map folding problem.
		pathLikeWriteFoldsTotal (getPathJobRootDEFAULT): Path, filename, or relative path and filename.
			If None, uses default path. If a directory, appends standardized filename.

	Returns:
		pathFilenameFoldsTotal: Absolute path and filename for storing the foldsTotal value.

	Notes:
		The function creates any necessary directories in the path if they don't exist.
	"""

	if pathLikeWriteFoldsTotal is None:
		pathFilenameFoldsTotal = getPathRootJobDEFAULT() / getFilenameFoldsTotal(mapShape)
	else:
		pathLikeSherpa = Path(pathLikeWriteFoldsTotal)
		if pathLikeSherpa.is_dir():
			pathFilenameFoldsTotal = pathLikeSherpa / getFilenameFoldsTotal(mapShape)
		elif pathLikeSherpa.is_file() and pathLikeSherpa.is_absolute():
			pathFilenameFoldsTotal = pathLikeSherpa
		else:
			pathFilenameFoldsTotal = getPathRootJobDEFAULT() / pathLikeSherpa

	pathFilenameFoldsTotal.parent.mkdir(parents=True, exist_ok=True)
	return pathFilenameFoldsTotal

# TODO learn how to see this from the user's perspective
def getPathRootJobDEFAULT() -> Path:
	pathJobDEFAULT = The.pathPackage / "jobs"
	if 'google.colab' in sysModules:
		pathJobDEFAULT = Path("/content/drive/MyDrive") / "jobs"
	pathJobDEFAULT.mkdir(parents=True, exist_ok=True)
	return pathJobDEFAULT

def _saveFoldsTotal(pathFilename: PathLike[str] | PurePath, foldsTotal: int) -> None:
	pathFilenameFoldsTotal = Path(pathFilename)
	pathFilenameFoldsTotal.parent.mkdir(parents=True, exist_ok=True)
	pathFilenameFoldsTotal.write_text(str(foldsTotal))

def saveFoldsTotal(pathFilename: PathLike[str] | PurePath, foldsTotal: int) -> None:
	"""
	Save `foldsTotal` value to disk with multiple fallback mechanisms.

	This function attempts to save the computed `foldsTotal` value to the specified
	location, with backup strategies in case the primary save attempt fails.
	The robustness is critical since these computations may take days to complete.

	Parameters:
		pathFilename: Target save location for the `foldsTotal` value
		foldsTotal: The computed value to save

	Notes:
		If the primary save fails, the function will attempt alternative save methods:
		1. Print the value prominently to stdout
		2. Create a fallback file in the current working directory
		3. As a last resort, simply print the value
	"""
	try:
		_saveFoldsTotal(pathFilename, foldsTotal)
	except Exception as ERRORmessage:
		try:
			print(f"\nfoldsTotal foldsTotal foldsTotal foldsTotal foldsTotal\n\n{foldsTotal=}\n\nfoldsTotal foldsTotal foldsTotal foldsTotal foldsTotal\n")
			print(ERRORmessage)
			print(f"\nfoldsTotal foldsTotal foldsTotal foldsTotal foldsTotal\n\n{foldsTotal=}\n\nfoldsTotal foldsTotal foldsTotal foldsTotal foldsTotal\n")
			randomnessPlanB = (int(str(foldsTotal).strip()[-1]) + 1) * ['YO_']
			filenameInfixUnique = ''.join(randomnessPlanB)
			pathFilenamePlanB = os.path.join(os.getcwd(), 'foldsTotal' + filenameInfixUnique + '.txt')
			writeStreamFallback = open(pathFilenamePlanB, 'w')
			writeStreamFallback.write(str(foldsTotal))
			writeStreamFallback.close()
			print(str(pathFilenamePlanB))
		except Exception:
			print(foldsTotal)
	return None

def saveFoldsTotalFAILearly(pathFilename: PathLike[str] | PurePath) -> None:
	if Path(pathFilename).exists():
		raise FileExistsError(f"{pathFilename=} exists: a battle of overwriting might cause tears.")
	if not Path(pathFilename).parent.exists():
		raise FileNotFoundError(f"I received {pathFilename=} 0.000139 seconds ago from a function that promised it created the parent directory, but the parent directory does not exist. Fix that now, so your computation doesn't get deleted later. And be compassionate to others.")
	foldsTotal = 149302889205120
	_saveFoldsTotal(pathFilename, foldsTotal)
	if not Path(pathFilename).exists():
		raise FileNotFoundError(f"I just wrote a test file to {pathFilename=}, but it does not exist. Fix that now, so your computation doesn't get deleted later. And continually improve your empathy skills.")
	foldsTotalRead = int(Path(pathFilename).read_text())
	if foldsTotalRead != foldsTotal:
		raise FileNotFoundError(f"I wrote a test file to {pathFilename=} with contents of {str(foldsTotal)=}, but I read {foldsTotalRead=} from the file. Python says the values are not equal. Fix that now, so your computation doesn't get corrupted later. And be pro-social.")

def writeStringToHere(this: str, pathFilename: PathLike[str] | PurePath) -> None:
	"""
	Write a string to a file, creating parent directories if needed.

	Parameters:
		this: The string content to write to the file
		pathFilename: The target file path where the string should be written
	"""
	pathFilename = Path(pathFilename)
	pathFilename.parent.mkdir(parents=True, exist_ok=True)
	pathFilename.write_text(str(this))
	return None
